#include "qvnc.h"

QVnc::QVnc(QQuickItem *parent) :
    QQuickFramebufferObject(parent), m_host(""), m_port(0)
{
    setAcceptedMouseButtons(Qt::AllButtons);
    setAcceptHoverEvents(true);
    // this makes sure the rendering is called only when we explicitly ask for it
    setTextureFollowsItemSize(false);

    connect(this, &QVnc::windowChanged, this, &QVnc::winChanged);

    m_uploader = new Uploader;
    connect(m_uploader, &Uploader::uploadFinished, this, &QVnc::update);
    connect(&m_worker, &QThread::finished, m_uploader, &QObject::deleteLater);

    m_dispatcher = new Dispatcher(m_uploader);
    connect(&m_worker, &QThread::finished, m_dispatcher, &QObject::deleteLater);
    connect(m_uploader, &Uploader::uploadFinished, m_dispatcher, &Dispatcher::refresh);

    m_uploader->moveToThread(&m_worker);
    m_dispatcher->moveToThread(&m_worker);

    m_worker.start();

    connect(this, &QVnc::open_connection, m_dispatcher, &Dispatcher::open);
    connect(this, &QVnc::terminate_connection, m_dispatcher, &Dispatcher::terminate);
    connect(m_dispatcher, &Dispatcher::error, this, &QVnc::dispatch_error);
}

QVnc::~QVnc()
{
}

void QVnc::open()
{
    emit open_connection(this->m_host, this->m_port, (int) this->width(), (int) this->height());
}

void QVnc::terminate()
{
    emit terminate_connection();
}

void QVnc::setHost(QString host)
{
    if (m_host == host)
        return;

    m_host = host;
    emit hostChanged(host);
}

void QVnc::setPort(int port)
{
    if (m_port == port)
        return;

    m_port = port;
    emit portChanged(port);
}

void QVnc::dispatch_error()
{
    qDebug() << "error";
}

void QVnc::winChanged(QQuickWindow *window)
{
    if(window != NULL)
        connect(window, &QQuickWindow::openglContextCreated, m_uploader, &Uploader::createContext);
}

QQuickFramebufferObject::Renderer *QVnc::createRenderer() const
{
    auto renderer = new VncRenderer(window(),
                                    m_uploader->fillBuffer(),
                                    m_uploader->copyBuffer(),
                                    m_uploader->bitmapBuffer(),
                                    m_uploader->ycbcrBuffer());
    connect(renderer, &VncRenderer::rendered, m_dispatcher, &Dispatcher::decode);
    return renderer;
}

QString QVnc::host() const
{
    return m_host;
}

int QVnc::port() const
{
    return m_port;
}

void QVnc::mousePressEvent(QMouseEvent *event)
{
    Qt::MouseButtons buttons = event->buttons();
    m_lastMouseEvent = event->timestamp();
    if(m_dispatcher->mouseEvent(event->x(),
                                event->y(),
                                buttons & Qt::LeftButton,
                                buttons & Qt::MiddleButton,
                                buttons & Qt::RightButton,
                                false,
                                false))
        event->accept();
    else
        event->ignore();
}
void QVnc::mouseReleaseEvent(QMouseEvent *event)
{
    Qt::MouseButtons buttons = event->buttons();
    m_lastMouseEvent = event->timestamp();
    if(m_dispatcher->mouseEvent(event->x(),
                                event->y(),
                                buttons & Qt::LeftButton,
                                buttons & Qt::MiddleButton,
                                buttons & Qt::RightButton,
                                false,
                                false))
        event->accept();
    else
        event->ignore();
}

void QVnc::wheelEvent(QWheelEvent *event)
{
    Qt::MouseButtons buttons = event->buttons();
    QPoint rotated = event->angleDelta();
    m_lastMouseEvent = event->timestamp();
    if(m_dispatcher->mouseEvent(event->x(),
                                event->y(),
                                buttons & Qt::LeftButton,
                                buttons & Qt::MiddleButton,
                                buttons & Qt::RightButton,
                                rotated.x() > 0,
                                rotated.x() < 0))
        event->accept();
    else
        event->ignore();
}

void QVnc::mouseMoveEvent(QMouseEvent *event)
{
    // Filter mouse move events, we don't need 100+ of them in a second
    //if((event->timestamp() - m_lastMouseEvent) > 33) {
        Qt::MouseButtons buttons = event->buttons();
    //    m_lastMouseEvent = event->timestamp();
        if(m_dispatcher->mouseEvent(event->x(),
                                    event->y(),
                                    buttons & Qt::LeftButton,
                                    buttons & Qt::MiddleButton,
                                    buttons & Qt::RightButton,
                                    false,
                                    false))
            event->accept();
        else
            event->ignore();
    //}
}

void QVnc::hoverMoveEvent(QHoverEvent *event)
{
    // Qt bug: timestamp is zero for QHoverEvent
    //if((event->timestamp() - m_lastMouseEvent) > 33) {
    //    m_lastMouseEvent = event->timestamp();
        QPoint position = event->pos();
        if(m_dispatcher->mouseEvent(position.x(),
                                    position.y(),
                                    false,
                                    false,
                                    false,
                                    false,
                                    false))
            event->accept();
        else
            event->ignore();
    //}
}

Uploader *QVnc::uploader() const
{
    return m_uploader;
}

void QVnc::keyPressEvent(QKeyEvent *event)
{
}

void QVnc::keyReleaseEvent(QKeyEvent *event)
{
}
