#include "uploader.h"

#include <cstring>

#include <texturefactory.h>

#include "jpegdecoder.h"

#include <QElapsedTimer>

//QElapsedTimer timer;

Uploader::Uploader(QObject *parent) :
    QObject(parent), m_context(this), m_fillIndex(0), m_copyIndex(0), m_bitmapIndex(0), m_ycbcrIndex(0), m_jpegs(0), m_finished(false), m_first(true)
{
    // "applications must ensure that create() is only called on the main (GUI) thread"
    QSurfaceFormat format;
    format.setAlphaBufferSize(8);
    format.setRedBufferSize(8);
    format.setGreenBufferSize(8);
    format.setBlueBufferSize(8);
    format.setDepthBufferSize(0);
    format.setStencilBufferSize(0);
    format.setMajorVersion(2);
    format.setMinorVersion(0);
    format.setRenderableType(QSurfaceFormat::OpenGLES);
    format.setSwapBehavior(QSurfaceFormat::SingleBuffer);
    format.setSwapInterval(0);
    m_surface.setFormat(format);
    m_surface.create();

    for(std::size_t i = 0; i < m_workers.size(); ++i)
    {
        m_decoders[i] = new JpegDecoder(&m_queue);
        m_decoders[i]->moveToThread(&m_workers[i]);
        connect(&m_workers[i], &QThread::finished, m_decoders[i], &QObject::deleteLater);
        connect(m_decoders[i], &JpegDecoder::finishedJpeg, this, &Uploader::finishedJpeg);
        connect(this, &Uploader::operate, m_decoders[i], &JpegDecoder::operate);
        m_workers[i].start();
    }

    emit operate();
}

Uploader::~Uploader()
{
}

void Uploader::handleDebugMessage(const QOpenGLDebugMessage &debugMessage)
{
    qDebug() << debugMessage;
}

void Uploader::createContext(QOpenGLContext *context)
{
    m_context.setShareContext(context);
    m_context.create();
    m_context.makeCurrent(&m_surface);

#ifdef OGL_DEBUG
    QOpenGLDebugLogger *logger = new QOpenGLDebugLogger(this);
    connect(logger, &QOpenGLDebugLogger::messageLogged, this, &Uploader::handleDebugMessage, Qt::DirectConnection);
    logger->initialize();
    logger->startLogging();
#endif

    m_gl = m_context.functions();

    m_fillBuffer.setUsagePattern(QOpenGLBuffer::DynamicDraw);
    m_fillBuffer.create();
    m_fillBuffer.bind();
    m_fillBuffer.allocate(VERTEX_BUFFER * sizeof(Fill));

    m_copyBuffer.setUsagePattern(QOpenGLBuffer::DynamicDraw);
    m_copyBuffer.create();
    m_copyBuffer.bind();
    m_copyBuffer.allocate(VERTEX_BUFFER * sizeof(Copy));

    m_bitmapBuffer.setUsagePattern(QOpenGLBuffer::DynamicDraw);
    m_bitmapBuffer.create();
    m_bitmapBuffer.bind();
    m_bitmapBuffer.allocate(VERTEX_BUFFER * sizeof(Bitmap));

    m_ycbcrBuffer.setUsagePattern(QOpenGLBuffer::DynamicDraw);
    m_ycbcrBuffer.create();
    m_ycbcrBuffer.bind();
    m_ycbcrBuffer.allocate(VERTEX_BUFFER * sizeof(Bitmap));

    auto textureFactory = TextureFactory::create();
    m_rgbTexture = textureFactory->createTexture(QSize(TEXTURE_WIDTH, TEXTURE_HEIGHT));
    m_yTexture = textureFactory->createTexture(QSize(TEXTURE_WIDTH / 4, TEXTURE_HEIGHT));
    m_cbTexture = textureFactory->createTexture(QSize(TEXTURE_WIDTH / 8, TEXTURE_HEIGHT));
    m_crTexture = textureFactory->createTexture(QSize(TEXTURE_WIDTH / 8, TEXTURE_HEIGHT));

    Jpeg::y_width = m_yTexture->allocated.width() * 4;
    Jpeg::cb_width = m_cbTexture->allocated.width() * 4;
    Jpeg::cr_width = m_crTexture->allocated.width() * 4;

    m_gl->glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
}

void Uploader::gotFill(const GLushort x, const GLushort y,
                       const GLushort width, const GLushort height,
                       const GLubyte red, const GLubyte green, const GLubyte blue)
{
    if(m_fillIndex + 4 < VERTEX_BUFFER) {
        m_fillPointer[m_fillIndex++] = Fill({x,                      (GLushort) (y + height), red, green, blue, 0xFF});
        m_fillPointer[m_fillIndex++] = Fill({x,                      y,                       red, green, blue, 0xFF});
        m_fillPointer[m_fillIndex++] = Fill({(GLushort) (x + width), (GLushort) (y + height), red, green, blue, 0xFF});
        m_fillPointer[m_fillIndex++] = Fill({(GLushort) (x + width), y,                       red, green, blue, 0xFF});
    } else {
        qDebug() << "Out of memory";
    }
}

void Uploader::gotCopy(const GLushort src_x, const GLushort src_y,
                       const GLushort width, const GLushort height,
                       const GLushort dest_x, const GLushort dest_y)
{
    if(m_copyIndex + 4 < VERTEX_BUFFER) {
        m_copyPointer[m_copyIndex++] = Copy({src_x,                      (GLushort) (src_y + height), dest_x,                      (GLushort) (dest_y + height)});
        m_copyPointer[m_copyIndex++] = Copy({src_x,                      src_y,                       dest_x,                      dest_y});
        m_copyPointer[m_copyIndex++] = Copy({(GLushort) (src_x + width), (GLushort) (src_y + height), (GLushort) (dest_x + width), (GLushort) (dest_y + height)});
        m_copyPointer[m_copyIndex++] = Copy({(GLushort) (src_x + width), src_y,                       (GLushort) (dest_x + width), dest_y});
    } else {
        qDebug() << "Out of memory";
    }

    if(m_bitmapIndex + 4 < VERTEX_BUFFER) {
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({dest_x,                      (GLushort) (dest_y + height)});
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({dest_x,                      dest_y});
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({(GLushort) (dest_x + width), (GLushort) (dest_y + height)});
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({(GLushort) (dest_x + width), dest_y});
    } else {
        qDebug() << "Out of memory";
    }
}

void Uploader::gotBitmap(const unsigned char *image,
                         const GLushort x, const GLushort y,
                         const GLushort width, const GLushort height)
{
    for (std::size_t row = 0; row < height; ++row)
        std::memcpy(&m_rgbPointer[(x + (y + row) * m_rgbTexture->allocated.width()) * 4], &image[row * width * 4], width * 4);

    if(m_bitmapIndex + 4 < VERTEX_BUFFER) {
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({x,                      (GLushort) (y + height)});
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({x,                      y});
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({(GLushort) (x + width), (GLushort) (y + height)});
        m_bitmapPointer[m_bitmapIndex++] = Bitmap({(GLushort) (x + width), y});
    } else {
        qDebug() << "Out of memory";
    }
}

void Uploader::gotJpeg(const uint8_t *buffer, int length, int x, int y, int width, int height)
{
    Jpeg *jpeg = new Jpeg((unsigned char*)buffer, length, width, height, x, y);
    m_queue.enqueue(jpeg);

    if(m_ycbcrIndex + 4 < VERTEX_BUFFER) {
        m_ycbcrPointer[m_ycbcrIndex++] = Bitmap({static_cast<GLushort>(x),         static_cast<GLushort>(y + height)});
        m_ycbcrPointer[m_ycbcrIndex++] = Bitmap({static_cast<GLushort>(x),         static_cast<GLushort>(y)});
        m_ycbcrPointer[m_ycbcrIndex++] = Bitmap({static_cast<GLushort>(x + width), static_cast<GLushort>(y + height)});
        m_ycbcrPointer[m_ycbcrIndex++] = Bitmap({static_cast<GLushort>(x + width), static_cast<GLushort>(y)});
    } else {
        qDebug() << "Out of memory";
    }

    m_jpegs++;
}

void Uploader::finishedJpeg()
{
    //qDebug() << "jpeg" << timer.elapsed();
    m_jpegs--;
    if(m_jpegs == 0 && m_finished) {
        startRendering();
    }
}

void Uploader::finishedUpdate()
{
    //qDebug() << "finished" << timer.elapsed();
    m_finished = true;
    if(m_jpegs == 0)
        startRendering();
}

void Uploader::startDecoding()
{
    //timer.start();
    if (m_fillIndex > 0 || m_first)
    {
        m_fillBuffer.bind();
        m_fillPointer = (Fill*) m_fillBuffer.map(QOpenGLBuffer::WriteOnly);
    }
    if (m_copyIndex > 0 || m_first)
    {
        m_copyBuffer.bind();
        m_copyPointer = (Copy*) m_copyBuffer.map(QOpenGLBuffer::WriteOnly);
    }
    if (m_bitmapIndex > 0 || m_first)
    {
        m_bitmapBuffer.bind();
        m_bitmapPointer = (Bitmap*) m_bitmapBuffer.map(QOpenGLBuffer::WriteOnly);
        m_rgbPointer = static_cast<unsigned char*>(m_rgbTexture->map());
    }
    if (m_ycbcrIndex > 0 || m_first)
    {
        m_ycbcrBuffer.bind();
        m_ycbcrPointer = (Bitmap*) m_ycbcrBuffer.map(QOpenGLBuffer::WriteOnly);
        Jpeg::y_pointer = static_cast<unsigned char*>(m_yTexture->map());
        Jpeg::cb_pointer = static_cast<unsigned char*>(m_cbTexture->map());
        Jpeg::cr_pointer = static_cast<unsigned char*>(m_crTexture->map());
    }

    m_fillIndex = 0;
    m_copyIndex = 0;
    m_bitmapIndex = 0;
    m_ycbcrIndex = 0;

    m_first = false;
    m_finished = false;
    m_jpegs = 0;
}

void Uploader::startRendering()
{
    if (m_fillIndex > 0)
    {
        m_fillBuffer.bind();
        m_fillBuffer.unmap();
    }
    if (m_copyIndex > 0)
    {
        m_copyBuffer.bind();
        m_copyBuffer.unmap();
    }
    if (m_bitmapIndex > 0)
    {
        m_bitmapBuffer.bind();
        m_bitmapBuffer.unmap();
        m_rgbTexture->unmap();
    }
    if (m_ycbcrIndex > 0)
    {
        m_ycbcrBuffer.bind();
        m_ycbcrBuffer.unmap();
        m_yTexture->unmap();
        m_cbTexture->unmap();
        m_crTexture->unmap();
    }
    //qDebug() << "uploaded" << timer.elapsed();

    emit uploadFinished();
}
