/*
 * Design decisions:
 * - There is no way to access texture memory directly, copying is required.
 *   See: https://www.raspberrypi.org/forums/viewtopic.php?f=67&t=25959
 * - libvnc hooks free the memory buffer after they return (except for gotJpeg)
 *   Solutions:
 *    1) copy images to separate textures
 *       Pro: simple
 *       Con: multiple textures (slow)
 *    2) make deep copies of the images and move them to a single texture when decoding finished
 *       Pro: optimal texture atlas
 *       Con: deep copy
 *    3) first-fit allocation
 */

#ifndef UPLOADER_H
#define UPLOADER_H

#define TEXTURE_WIDTH 1920
#define TEXTURE_HEIGHT 1080
#define VERTEX_BUFFER 16384
//#define OGL_DEBUG

#include <utility>
#include <memory>

#include <QObject>
#include <QDebug>
#include <QOffscreenSurface>
#include <QOpenGLContext>
#include <QOpenGLFunctions>
#include <QOpenGLBuffer>
#include <QOpenGLShaderProgram>
#include <QtMath>
#include <QTime>
#include <QOpenGLDebugLogger>
#include <QThread>

#include <texture.h>

#include "concurrentqueue.h"

#define DECODER_COUNT 5

struct Bitmap
{
    GLushort x;
    GLushort y;
};

struct Copy
{
    GLushort src_x;
    GLushort src_y;
    GLushort dest_x;
    GLushort dest_y;
};

struct Fill
{
    GLushort x;
    GLushort y;
    GLubyte red;
    GLubyte green;
    GLubyte blue;
    GLubyte alpha;
};

class JpegDecoder;
class Jpeg;

/*!
 * \brief The Uploader class
 *
 * A framebuffer is bound to the
 */
class Uploader : public QObject
{
    Q_OBJECT
public:
    explicit Uploader(QObject *parent = Q_NULLPTR);
    ~Uploader();

    // might as well make these public, they are shared with the renderer
    std::shared_ptr<Texture> rgbTexture() const {return m_rgbTexture;}
    std::shared_ptr<Texture> yTexture() const {return m_yTexture;}
    std::shared_ptr<Texture> cbTexture() const {return m_cbTexture;}
    std::shared_ptr<Texture> crTexture() const {return m_crTexture;}
    QOpenGLBuffer fillBuffer() const {return m_fillBuffer;}
    QOpenGLBuffer copyBuffer() const {return m_copyBuffer;}
    QOpenGLBuffer bitmapBuffer() const {return m_bitmapBuffer;}
    QOpenGLBuffer ycbcrBuffer() const {return m_ycbcrBuffer;}
    std::size_t fillCount() const {return m_fillIndex;}
    std::size_t copyCount() const {return m_copyIndex;}
    std::size_t bitmapCount() const {return m_bitmapIndex;}
    std::size_t ycbcrCount() const {return m_ycbcrIndex;}

    // VNC events
    void gotFill(const GLushort x, const GLushort y,
                 const GLushort width, const GLushort height,
                 const GLubyte red, const GLubyte green, const GLubyte blue);
    void gotCopy(const GLushort src_x, const GLushort src_y,
                 const GLushort width, const GLushort height,
                 const GLushort dest_x, const GLushort dest_y);
    void gotBitmap(const unsigned char *image,
                   const GLushort x, const GLushort y,
                   const GLushort width, const GLushort height);
    void gotJpeg(const uint8_t *buffer, int length, int x, int y, int width, int height);
    void finishedUpdate();
    void startDecoding();

private:
    QOffscreenSurface m_surface;
    QOpenGLContext m_context;
    QOpenGLFunctions *m_gl;

    QOpenGLBuffer m_fillBuffer;
    Fill *m_fillPointer;
    std::size_t m_fillIndex;

    QOpenGLBuffer m_copyBuffer;
    Copy *m_copyPointer;
    std::size_t m_copyIndex;

    QOpenGLBuffer m_bitmapBuffer;
    Bitmap *m_bitmapPointer;
    std::size_t m_bitmapIndex;

    std::shared_ptr<Texture> m_rgbTexture;
    unsigned char *m_rgbPointer;

    QOpenGLBuffer m_ycbcrBuffer;
    Bitmap *m_ycbcrPointer;
    std::size_t m_ycbcrIndex;

    std::shared_ptr<Texture> m_yTexture;
    std::shared_ptr<Texture> m_cbTexture;
    std::shared_ptr<Texture> m_crTexture;

    std::array<QThread, DECODER_COUNT> m_workers;
    std::array<JpegDecoder*, DECODER_COUNT> m_decoders;
    ConcurrentQueue<Jpeg*, 1024u> m_queue;

    int m_jpegs;
    bool m_finished;

    bool m_first;

    void startRendering();
signals:
    void uploadFinished();
    void buffersSwapped();
    void operate();
public slots:
    void createContext(QOpenGLContext *context);
    void finishedJpeg();
private slots:
    void handleDebugMessage(const QOpenGLDebugMessage &debugMessage);
};

#endif // UPLOADER_H
