#include "vncrenderer.h"
#include <cstring>
#include "uploader.h"

VncRenderer::VncRenderer(QQuickWindow *window,
                         QOpenGLBuffer fillBuffer,
                         QOpenGLBuffer copyBuffer,
                         QOpenGLBuffer bitmapBuffer,
                         QOpenGLBuffer ycbcrBuffer,
                         QObject *parent) :
    QObject(parent), m_window(window),
    m_fillBuffer(fillBuffer), m_copyBuffer(copyBuffer), m_bitmapBuffer(bitmapBuffer), m_ycbcrBuffer(ycbcrBuffer),
    m_fillCount(0), m_copyCount(0), m_bitmapCount(0), m_ycbcrCount(0),
    m_indices(QOpenGLBuffer::IndexBuffer)
{
#ifdef OGL_DEBUG
    QOpenGLDebugLogger *logger = new QOpenGLDebugLogger(this);
    connect(logger, &QOpenGLDebugLogger::messageLogged, this, &VncRenderer::handleDebugMessage, Qt::DirectConnection);
    logger->initialize();
    logger->startLogging();
#endif

    m_toTexture.ortho(QRect(QPoint(0, TEXTURE_HEIGHT), QSize(TEXTURE_WIDTH, -1 * TEXTURE_HEIGHT)));
    m_fromTexture.ortho(QRect(QPoint(-1 * TEXTURE_WIDTH, TEXTURE_HEIGHT), QSize(2 * TEXTURE_WIDTH, -2 * TEXTURE_HEIGHT)));

    m_fillProgram.addShaderFromSourceFile(QOpenGLShader::Vertex, "/home/ldmnyblzs/Projects/ThinClient/fill_shader.vsh");
    m_fillProgram.addShaderFromSourceFile(QOpenGLShader::Fragment, "/home/ldmnyblzs/Projects/ThinClient/fill_shader.fsh");
    m_fillProgram.link();

    m_copyProgram.addShaderFromSourceFile(QOpenGLShader::Vertex, "/home/ldmnyblzs/Projects/ThinClient/copy_shader.vsh");
    m_copyProgram.addShaderFromSourceFile(QOpenGLShader::Fragment, "/home/ldmnyblzs/Projects/ThinClient/copy_shader.fsh");
    m_copyProgram.bind();
    m_copyProgram.setUniformValue("trgortho", m_toTexture);
    m_copyProgram.setUniformValue("texture", 0);
    m_copyProgram.release();

    m_bitmapProgram.addShaderFromSourceFile(QOpenGLShader::Vertex, "/home/ldmnyblzs/Projects/ThinClient/bitmap_shader.vsh");
    m_bitmapProgram.addShaderFromSourceFile(QOpenGLShader::Fragment, "/home/ldmnyblzs/Projects/ThinClient/bitmap_shader.fsh");

    m_bitmapProgram.bind();
    m_bitmapProgram.setUniformValue("texture", 1);
    m_bitmapProgram.setUniformValue("texortho", m_fromTexture);
    m_bitmapProgram.release();

    m_ycbcrProgram.addShaderFromSourceFile(QOpenGLShader::Vertex, "/home/ldmnyblzs/Projects/ThinClient/bitmap_shader.vsh");
    m_ycbcrProgram.addShaderFromSourceFile(QOpenGLShader::Fragment, "/home/ldmnyblzs/Projects/ThinClient/ycbcr_shader.fsh");

    m_ycbcrProgram.bind();
    m_ycbcrProgram.setUniformValue("y_texture", 2);
    m_ycbcrProgram.setUniformValue("cb_texture", 3);
    m_ycbcrProgram.setUniformValue("cr_texture", 4);
    m_ycbcrProgram.setUniformValue("texortho", m_fromTexture);
    m_ycbcrProgram.release();

    m_indices.create();
    m_indices.bind();
    m_indices.allocate(ind.i, (VERTEX_BUFFER / 4) * 6 - 2);
    m_indices.release();

    m_gl = QOpenGLContext::currentContext()->functions();

    m_gl->glGenFramebuffers(1, &m_textureTarget);

    m_clear = true;
    frameCount = 0;
    frameTime.start();
}

void VncRenderer::handleDebugMessage(const QOpenGLDebugMessage &debugMessage)
{
    qDebug() << debugMessage;
}

void VncRenderer::render()
{
    if (m_clear)
    {
        m_gl->glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
        m_gl->glClear(GL_COLOR_BUFFER_BIT);
        m_clear = false;
    }

    if(m_fillCount > 0 || m_bitmapCount > 0 || m_ycbcrCount > 0)
    {
        //frameCount++;
    }
    else
        return;

    /*if(frameTime.elapsed() > 1000) {
        qDebug() << frameCount;
        frameCount = 0;
        frameTime.restart();
    }*/

    /* Rendering:
     * 1. Draw filled rectangles
     * 2. Copy rectangles from the framebuffer to a temporary texture
     * 3. Draw RGB textured rectangles
     * 4. Draw YCbCr textured rectangles
     */

    m_indices.bind();

    if(m_fillCount > 0) {
        m_fillProgram.bind();
        m_fillBuffer.bind();
        m_fillProgram.enableAttributeArray("position");
        m_fillProgram.enableAttributeArray("color");
        m_gl->glVertexAttribPointer(m_fillProgram.attributeLocation("position"),
                                    2,
                                    GL_UNSIGNED_SHORT,
                                    GL_FALSE,
                                    sizeof(Fill),
                                    NULL);
        m_gl->glVertexAttribPointer(m_fillProgram.attributeLocation("color"),
                                    4,
                                    GL_UNSIGNED_BYTE,
                                    GL_TRUE,
                                    sizeof(Fill),
                                    reinterpret_cast<void*>(2 * sizeof(GLushort)));

        m_gl->glDrawElements(GL_TRIANGLE_STRIP, m_fillCount, GL_UNSIGNED_SHORT, NULL);

        m_fillBuffer.release();
        m_fillProgram.release();
    }

    if(m_copyCount > 0) {
        m_gl->glActiveTexture(GL_TEXTURE0);
        m_gl->glBindTexture(GL_TEXTURE_2D, m_framebufferObject->texture());
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        m_copyProgram.bind();
        m_copyBuffer.bind();
        m_copyProgram.enableAttributeArray("source");
        m_copyProgram.enableAttributeArray("target");
        m_gl->glVertexAttribPointer(m_copyProgram.attributeLocation("source"),
                                    2,
                                    GL_UNSIGNED_SHORT,
                                    GL_FALSE,
                                    sizeof(Copy),
                                    NULL);
        m_gl->glVertexAttribPointer(m_copyProgram.attributeLocation("target"),
                                    2,
                                    GL_UNSIGNED_SHORT,
                                    GL_FALSE,
                                    sizeof(Copy),
                                    reinterpret_cast<void*>(2 * sizeof(GLushort)));

        m_gl->glBindFramebuffer(GL_FRAMEBUFFER, m_textureTarget);
        m_gl->glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, *m_rgbTexture, 0);
        m_gl->glViewport(0, 0, TEXTURE_WIDTH, TEXTURE_HEIGHT);

        m_gl->glDrawElements(GL_TRIANGLE_STRIP, m_copyCount, GL_UNSIGNED_SHORT, NULL);

        m_framebufferObject->bind();
        m_gl->glViewport(0, 0, 1920, 1080);

        m_copyBuffer.release();
        m_copyProgram.release();
    }

    if(m_bitmapCount > 0) {
        m_rgbTexture->bind(GL_TEXTURE1);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        m_bitmapProgram.bind();
        m_bitmapBuffer.bind();
        m_bitmapProgram.enableAttributeArray("position");
        m_gl->glVertexAttribPointer(m_bitmapProgram.attributeLocation("position"),
                                    2,
                                    GL_UNSIGNED_SHORT,
                                    GL_FALSE,
                                    0,
                                    NULL);
        m_gl->glDrawElements(GL_TRIANGLE_STRIP, m_bitmapCount, GL_UNSIGNED_SHORT, NULL);
        m_bitmapBuffer.release();
        m_bitmapProgram.release();
    }

    if(m_ycbcrCount > 0) {
        m_yTexture->bind(GL_TEXTURE2);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        m_crTexture->bind(GL_TEXTURE3);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        m_cbTexture->bind(GL_TEXTURE4);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        m_gl->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        m_ycbcrProgram.bind();
        m_ycbcrBuffer.bind();
        m_ycbcrProgram.enableAttributeArray("position");
        m_gl->glVertexAttribPointer(m_ycbcrProgram.attributeLocation("position"),
                                    2,
                                    GL_UNSIGNED_SHORT,
                                    GL_FALSE,
                                    0,
                                    NULL);
        m_gl->glDrawElements(GL_TRIANGLE_STRIP, m_ycbcrCount, GL_UNSIGNED_SHORT, NULL);
        m_ycbcrBuffer.release();
        m_ycbcrProgram.release();
    }

    m_indices.release();

    m_window->resetOpenGLState();

    emit rendered();
}

void VncRenderer::synchronize(QQuickFramebufferObject *fbo)
{
    auto qvnc = static_cast<QVnc*>(fbo);
    auto uploader = qvnc->uploader();
    m_fillCount = std::max(uploader->fillCount() / 4 * 6, 2uL) - 2;
    m_copyCount = std::max(uploader->copyCount() / 4 * 6, 2uL) - 2;
    m_bitmapCount = std::max(uploader->bitmapCount() / 4 * 6, 2uL) - 2;
    m_ycbcrCount = std::max(uploader->ycbcrCount() / 4 * 6, 2uL) - 2;

    // query until it is finally set
    if (m_rgbTexture.get() == nullptr)
    {
        m_rgbTexture = uploader->rgbTexture();
        m_yTexture = uploader->yTexture();
        m_cbTexture = uploader->cbTexture();
        m_crTexture = uploader->crTexture();
    }
}

QOpenGLFramebufferObject *VncRenderer::createFramebufferObject(const QSize &size)
{
    m_toFramebuffer.ortho(QRect(QPoint(0, size.height()), QSize(size.width(), -1 * size.height())));
    m_fromFramebuffer.ortho(QRect(QPoint(-1 * size.width(), size.height()), QSize(2 * size.width(), -2 * size.height())));

    m_fillProgram.bind();
    m_fillProgram.setUniformValue("ortho", m_toFramebuffer);
    m_fillProgram.release();

    m_copyProgram.bind();
    m_copyProgram.setUniformValue("srcortho", m_fromFramebuffer);
    m_copyProgram.release();

    m_bitmapProgram.bind();
    m_bitmapProgram.setUniformValue("ortho", m_toFramebuffer);
    m_bitmapProgram.release();

    m_ycbcrProgram.bind();
    m_ycbcrProgram.setUniformValue("ortho", m_toFramebuffer);
    m_ycbcrProgram.release();

    return (m_framebufferObject = new QOpenGLFramebufferObject(size));
}
